@@ -109,11 +109,26 @@ end
 
 -- set the value to the package info
 function _instance:set(name, ...)
+    if self._SOURCE_INITED then
+        -- we can use set/add to modify urls, .. in on_load() if urls have been inited.
+        -- but we cannot init urls, ... in on_load() if it has been not inited
+        --
+        -- @see https://github.com/xmake-io/xmake/issues/5148
+        -- https://github.com/xmake-io/xmake-repo/pull/4204
+        if self:_sourceset():has(name) and self:get(name) == nil then
+            os.raise("'%s' can only be initied in on_source() or the description scope.", name)
+        end
+    end
     self._INFO:apival_set(name, ...)
 end
 
 -- add the value to the package info
 function _instance:add(name, ...)
+    if self._SOURCE_INITED then
+        if self:_sourceset():has(name) and self:get(name) == nil then
+            os.raise("'%s' can only be initied in on_source() or the description scope.", name)
+        end
+    end
     self._INFO:apival_add(name, ...)
 end
 
@@ -889,6 +904,7 @@ function _instance:manifest_save()
     manifest.mode        = self:mode()
     manifest.configs     = self:configs()
     manifest.envs        = self:_rawenvs()
+    manifest.pathenvs    = self:_pathenvs():to_array()
 
     -- save enabled library deps
     if self:librarydeps() then
@@ -975,6 +991,40 @@ function _instance:manifest_save()
     end
 end
 
+-- get the source configuration set
+function _instance:_sourceset()
+    local sourceset = self._SOURCESET
+    if sourceset == nil then
+        sourceset = hashset.of("urls", "versions", "versionfiles", "configs")
+        self._SOURCESET = sourceset
+    end
+    return sourceset
+end
+
+-- init package source
+function _instance:_init_source()
+    local inited = self._SOURCE_INITED
+    if not inited then
+        local on_source = self:script("source")
+        if on_source then
+            on_source(self)
+        end
+    end
+end
+
+-- load package
+function _instance:_load()
+    self._SOURCE_INITED = true
+    local loaded = self._LOADED
+    if not loaded then
+        local on_load = self:script("load")
+        if on_load then
+            on_load(self)
+        end
+        self._LOADED = true
+    end
+end
+
 -- get the raw environments
 function _instance:_rawenvs()
     local envs = self._RAWENVS
@@ -1288,8 +1338,12 @@ function _instance:_versions_list()
         local versionfiles = self:get("versionfiles")
         if versionfiles then
             for _, versionfile in ipairs(table.wrap(versionfiles)) do
-                if not os.isfile(versionfile) then
-                    versionfile = path.join(self:scriptdir(), versionfile)
+                if not path.is_absolute(versionfile) then
+                    local subpath = versionfile
+                    versionfile = path.join(self:scriptdir(), subpath)
+                    if not os.isfile(versionfile) then
+                        versionfile = path.join(self:base():scriptdir(), subpath)
+                    end
                 end
                 if os.isfile(versionfile) then
                     local list = io.readfile(versionfile)
@@ -2661,7 +2715,8 @@ function package.apis()
     ,   script =
         {
             -- package.on_xxx
-            "package.on_load"
+            "package.on_source"
+        ,   "package.on_load"
         ,   "package.on_fetch"
         ,   "package.on_check"
         ,   "package.on_download"
